# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import TYPE_CHECKING, Any, Dict, List, Mapping, NoReturn, Optional, Sequence
import warnings

from ..damlast import IdentityTypeVisitor
from ..damlast.daml_lf_1 import DefDataType, FieldWithType, PrimType, Type
from ..damlast.lookup import EmptyLookup
from ..damlast.protocols import SymbolLookup
from ..prim import ContractId, DazlError, to_str, to_variant
from .mapper import ValueMapper

if TYPE_CHECKING:
    from .mapper import ValueMapper


__all__ = ["Context", "UnboundVarError"]


# Context is designed to be subclassable, so avoid making methods that could be static, static.
# noinspection PyMethodMayBeStatic
class Context:
    """
    A context under which object translation can occur.

    Usage:

    > context = Context(mapper)
    > context.convert(obj_type, obj)
    """

    def __init__(
        self,
        mapper: "ValueMapper",
        lookup: "Optional[SymbolLookup]" = None,
        path: "Sequence[str]" = (),
    ):
        """
        Initialize a Context.

        Note that ``None`` as a :class:`SymbolLookup` means that mappers that use the created
         :class:`Context` can only understand and map primitive types; this may be suitable for
        some limited use cases, so it is allowed.

        :param lookup:
            :class:`SymbolLookup` implementation that can provide references to types and templates.
            If ``None``, mappers that use this context cannot resolve any complex types.
        """
        self.mapper = mapper
        self.lookup = lookup if lookup is not None else EmptyLookup()
        self.path = tuple(path)

    def convert(self, item_type: "Type", obj: "Any"):
        """
        Convert a value from one representation to another.

        :param item_type:
            The assumed DAML-LF :class:`Type` of ``obj``.
        :param obj:
            The object to convert.
        """
        try:
            if item_type.var is not None:
                # We support type variables, but only when they are bound; if we receive a type
                # variable here (outside of a Type.Con context where type arguments can be
                # immediately applied and substituted) we can't meaningfully process them.
                raise UnboundVarError(item_type.var.var)
            elif item_type.con is not None:
                dt = self.resolve_data_type(item_type.con)
                if dt.record is not None:
                    return self.mapper.data_record(self, dt, dt.record, obj)
                elif dt.variant is not None:
                    return self.mapper.data_variant(self, dt, dt.variant, obj)
                elif dt.enum is not None:
                    return self.mapper.data_enum(self, dt, dt.enum, obj)
            elif item_type.prim is not None:
                if item_type.prim.prim == PrimType.UNIT:
                    return self.mapper.prim_unit(self, obj)
                elif item_type.prim.prim == PrimType.BOOL:
                    return self.mapper.prim_bool(self, obj)
                elif item_type.prim.prim == PrimType.INT64:
                    return self.mapper.prim_int64(self, obj)
                elif item_type.prim.prim == PrimType.DECIMAL:
                    return self.mapper.prim_numeric(self, 10, obj)
                elif item_type.prim.prim == PrimType.TEXT or item_type.prim.prim == PrimType.CHAR:
                    return self.mapper.prim_text(self, obj)
                elif item_type.prim.prim == PrimType.TIMESTAMP:
                    return self.mapper.prim_timestamp(self, obj)
                elif item_type.prim.prim == PrimType.PARTY:
                    return self.mapper.prim_party(self, obj)
                elif item_type.prim.prim == PrimType.LIST:
                    return self.mapper.prim_list(self, item_type.prim.args[0], obj)
                elif item_type.prim.prim == PrimType.DATE:
                    return self.mapper.prim_date(self, obj)
                elif item_type.prim.prim == PrimType.CONTRACT_ID:
                    return self.mapper.prim_contract_id(self, item_type.prim.args[0], obj)
                elif item_type.prim.prim == PrimType.OPTIONAL:
                    return self.mapper.prim_optional(self, item_type.prim.args[0], obj)
                elif item_type.prim.prim == PrimType.TEXTMAP:
                    return self.mapper.prim_text_map(self, item_type.prim.args[0], obj)
                elif item_type.prim.prim == PrimType.NUMERIC:
                    return self.mapper.prim_numeric(self, item_type.prim.args[0].nat, obj)
                elif item_type.prim.prim == PrimType.GENMAP:
                    return self.mapper.prim_gen_map(
                        self, item_type.prim.args[0], item_type.prim.args[1], obj
                    )

            raise ValueError(f"could not process type: {item_type}")

        except ValueError as ex:
            # re-raise ValueErrors with a more descriptive string that includes the current context
            # path
            raise ValueError(f'value at {".".join(self.path)} has an invalid value: {obj}') from ex

    def convert_list(
        self,
        element_type: "Type",
        elements: "Sequence[Any]",
        mapper: "Optional[ValueMapper]" = None,
    ) -> "List[Any]":
        """
        Convert a _list_ of elements, each of an assumed type.

        This is a convenience method for :class:`ValueMapper` implementations where lists are
        implemented as Python lists. Note that a :class:`ValueMapper` may choose to implement list
        type conversion in a completely different way instead of calling this function, which may be
        necessary if a list of DAML-LF values is encoded in a way other than a Python list
        (i.e., Protobuf).

        :param element_type:
            The :class:`Type` of the element.
        :param elements:
            An iterable collection of objects to convert.
        :param mapper:
            If not ``None``, an alternate mapper to use when descending down the object.
        """
        return [
            self.append_path(f"[{i}", mapper=mapper).convert(element_type, elem)
            for i, elem in enumerate(elements)
        ]

    def convert_optional(
        self, element_type: "Type", element: "Optional[Any]", mapper: "Optional[ValueMapper]" = None
    ) -> "List[Any]":
        """
        Convert a _list_ of elements, each of an assumed type.

        This is a convenience method for :class:`ValueMapper` implementations where DAML-LF
        Optionals are implemented as normal Python values or ``None``. A :class:`ValueMapper` may
        choose to implement optional type conversion in a completely different way, which may be
        necessary if an Optional of a DAML-LF value is encoded in a way other than its natural
        representation or ``None`` (i.e., Protobuf).

        :param element_type:
            The :class:`Type` of the element.
        :param element:
            An object to convert.
        :param mapper:
            If not ``None``, an alternate mapper to use when descending down the object.
        :return:
            The converted object, or ``None`` if the value of the object is ``None``.
        """
        return (
            self.append_path("?", mapper=mapper).convert(element_type, element)
            if element is not None
            else None
        )

    def convert_text_map(
        self,
        element_type: "Type",
        elements: "Mapping[str, Any]",
        mapper: "Optional[ValueMapper]" = None,
    ) -> "Dict[str, Any]":
        """
        Convert a _map_ of elements, each of an assumed type.

        This is a convenience method for :class:`ValueMapper` implementations where DAML-LF TextMaps
        are represented as Python dictionaries. Note that a :class:`ValueMapper` may choose to
        implement TextMap type conversion in a completely different way instead of calling this
        function, which may be necessary if a TextMap of DAML-LF values is encoded in a way other
        than a Python dict (i.e., Protobuf).

        :param element_type:
            The :class:`Type` of the element.
        :param elements:
            A mapping of keys to values.
        :param mapper:
            If not ``None``, an alternate mapper to use when descending down the object.
        """
        return {
            key: self.append_path(key, mapper=mapper).convert(element_type, value)
            for key, value in elements.items()
        }

    def convert_gen_map(
        self,
        key_type: "Type",
        value_type: "Type",
        elements: "Mapping[Any, Any]",
        mapper: "Optional[ValueMapper]" = None,
    ) -> "Dict[str, Any]":
        """
        Convert a _map_ of elements, each of an assumed type.

        This is a convenience method for :class:`ValueMapper` implementations where DAML-LF TextMaps
        are represented as Python dictionaries. Note that a :class:`ValueMapper` may choose to
        implement TextMap type conversion in a completely different way instead of calling this
        function, which may be necessary if a TextMap of DAML-LF values is encoded in a way other
        than a Python dict (i.e., Protobuf).

        :param key_type:
            The :class:`Type` of the key.
        :param value_type:
            The :class:`Type` of the value.
        :param elements:
            A mapping of keys to values.
        :param mapper:
            If not ``None``, an alternate mapper to use when descending down the object.
        """
        return {
            self.append_path("[i]", mapper=mapper)
            .convert(key_type, key): self.append_path(key, mapper=mapper)
            .convert(value_type, value)
            for i, (key, value) in enumerate(elements.items())
        }

    def convert_contract_id(self, element_type: "Type", contract_id: "Any") -> "ContractId":
        """
        Convert a contract ID string to a :class:`ContractId`.
        """
        with warnings.catch_warnings():
            # TODO: Drop this and switch to new-style ContractIds
            warnings.simplefilter("ignore", DeprecationWarning)
            from ..model.core import ContractId as DeprecatedContractId

            if isinstance(contract_id, DeprecatedContractId):
                return contract_id
            elif isinstance(contract_id, ContractId):
                # convert new-style ContractId instances to deprecated subclasses; for now, this is
                # still the canonical form of a ContractId until the deprecated version is dropped
                return DeprecatedContractId(contract_id.value, element_type.con.tycon)
            else:
                return DeprecatedContractId(contract_id, element_type.con.tycon)

    def resolve_data_type(self, con: "Type.Con") -> "DefDataType":
        """
        Resolve a :class:`DefDataType`, including applying any type parameters.

        In order to support recursive types and type variables, only the variables that are fields
        of the resolved :class:`DefDataType` are replaced.
        """
        dt = self.lookup.data_type(con.tycon)
        if len(dt.params) != len(con.args):
            # every time we encounter DefDataType, we expect to be given its type arguments in the
            # same place; there are no other constructs in DAML-LF that express type application,
            # so failure to supply type arguments here is an error
            raise ValueError("partially-applied types are not allowed")

        # without type arguments, simply return the original DefDataType; it does not need to be
        # modified
        if not con.args:
            return dt

        # from this point on, we discard all type arguments and the original name of the DefDataType
        # since the fully-applied type is NOT the same type as the original, nor is it correct to
        # say it takes any type arguments

        type_vars = {tvwk.var: arg for (tvwk, arg) in zip(dt.params, con.args)}

        if dt.record is not None:
            return DefDataType(
                record=self._replace_all_type_vars(type_vars, dt.record),
                serializable=dt.serializable,
                location=dt.location,
            )

        elif dt.variant is not None:
            return DefDataType(
                variant=self._replace_all_type_vars(type_vars, dt.variant),
                serializable=dt.serializable,
                location=dt.location,
            )

        elif dt.enum is not None:
            return DefDataType(enum=dt.enum, serializable=dt.serializable, location=dt.location)

        elif dt.synonym is not None:
            return DefDataType(
                synonym=dt.synonym, serializable=dt.serializable, location=dt.location
            )

        else:
            raise ValueError("unknown DefDataType cannot have variables applied to them")

    def append_path(self, path: str, mapper: "Optional[ValueMapper]" = None) -> "Context":
        """
        Return a new :class:`Context` marked at a deeper path within an object hierarchy.

        :param path:
            The path to append to the current :class:`Context`'s path.
        :param mapper:
            If not ``None``, an alternate mapper to use when descending down the object.
        :return:
            A new :class:`Context` with a modified path and possibly a modified mapper.
        """
        return Context(
            mapper if mapper is not None else self.mapper, self.lookup, self.path + (path,)
        )

    def value_validate_enum(self, value: "Any", enum: "DefDataType.EnumConstructors") -> str:
        """
        Return either a confirmed valid value from the list of possible enums, or throw an error.

        :param value:
            The "natural" Python type for an enum (either a string or a variant object).
        :param enum:
            The enum constructors to validate against.
        :return:
            The string that is the valid constructor from the list of constructors to choose from.
        :raise ValueError:
            if the value could not be understood as one of the enum constructors.
        """
        from collections.abc import Mapping

        if isinstance(value, Mapping):
            ctor, _ = to_variant(value)
        else:
            ctor = to_str(value)

        if ctor in enum.constructors:
            return ctor

        self.value_error(value, f"expected one of {enum.constructors}")

    def value_warn(self, value: "Any", message: str) -> "None":
        """
        Raise a :class:`ValueWarning`, enriched with information in the current context.
        """
        warnings.warn(
            f'value at {".".join(self.path)} has a possibly invalid value: {value} ({message})',
            ValueWarning,
            stacklevel=2,
        )

    def value_error(self, value: "Any", message: str) -> "NoReturn":
        """
        Raise a :class:`ValueError`, enriched with information in the current context.
        """
        raise ValueError(
            f'value at {".".join(self.path)} has an invalid value: {value} ({message})'
        )

    @property
    def depth(self) -> int:
        """
        Return the number of components in the path of this context.
        """
        return len(self.path)

    def _replace_all_type_vars(
        self, type_vars: "Mapping[str, Type]", fields: "DefDataType.Fields"
    ) -> "DefDataType.Fields":
        """
        Substitute type variables from the given mapping into each of the provided fields.
        """
        tv = ReplacingTypeVisitor(type_vars)
        return DefDataType.Fields(
            [FieldWithType(f.field, tv.visit_type(f.type)) for f in fields.fields]
        )


class ReplacingTypeVisitor(IdentityTypeVisitor):
    def __init__(self, type_vars: "Mapping[str, Type]"):
        self.type_vars = type_vars

    def resolve_type(self, var: "str") -> "Optional[Type]":
        replacement = self.type_vars.get(var)
        if replacement is None:
            raise UnboundVarError(var)

        return replacement


class UnboundVarError(DazlError):
    def __init__(self, var):
        self.var = var


class ValueWarning(Warning):
    pass



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import asyncio
import concurrent.futures
from typing import TYPE_CHECKING, Callable, Optional, TypeVar

if TYPE_CHECKING:
    from . import Connection


Fn = TypeVar("Fn", bound=Callable)

__all__ = [
    "CallbackReturnWarning",
    "ProtocolWarning",
    "ConnectionClosedError",
    "_rewrite_exceptions",
    "_translate_exceptions",
    "_allow_cancel",
]


class CallbackReturnWarning(Warning):
    """
    Raised when a user callback on a stream returns a value. These objects have no meaning and are
    ignored by dazl.

    This warning is raised primarily because older versions of dazl interpreted returning commands
    from a callback as a request to send commands to the underlying ledger, and this is not
    supported in newer APIs.
    """


class ProtocolWarning(Warning):
    """
    Warnings that are raised when dazl detects incompatibilities between the Ledger API server-side
    implementation and dazl.
    """


class ConnectionClosedError(Exception):
    """
    Raised when trying to call methods on a closed :class:`dazl.ledger.Connection`.
    """


def _rewrite_exceptions(fn: Fn) -> Fn:
    """
    Translate and re-raised internal gRPC exceptions to dazl's own error hierarchy. This
    decorator expects to wrap methods off of :class:`dazl.ledger.Connection` objects, or
    objects with a ``conn`` property (usually QueryStreams).

    Unfortunately gRPC does not have a standardized set of well-defined exceptions.
    Additionally the async API and the non-async API throw exceptions from different class
    hierarchies!
    """

    # Unfortunately the error types raised by the gRPC async API are not actually public
    # symbols...if a new version of the gRPC library changes this, then we'll fall back
    # to handling all Exceptions until this code is modified to handle this case.
    try:
        # noinspection PyProtectedMember
        from grpc._cython.cygrpc import UsageError  # type: ignore
    except ImportError:
        # noinspection PyPep8Naming
        UsageError = Exception

    def _rewrite_exception(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except UsageError:
            conn = getattr(self, "conn", self)
            if conn.is_closed:
                raise ConnectionClosedError()
            else:
                raise

    return _rewrite_exception  # type: ignore


class ExceptionTranslator:
    def __init__(self, conn: "Connection"):
        self.conn = conn

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None and self.conn.is_closed:
            raise ConnectionClosedError() from exc_val

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None and self.conn.is_closed:
            raise ConnectionClosedError() from exc_val


class AllowCancellation:
    def __init__(self, allow: Callable[[], bool]):
        self.allow = allow

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        if exc_val is not None:
            # These two exceptions are actually the same type under the hood--however, nothing
            # that I can find in the Python docs suggest that this HAS to be the case. Either
            # way, suppress cancellation errors that happen when we are closed, because .
            return self.allow() and (
                exc_type is asyncio.CancelledError or exc_type is concurrent.futures.CancelledError
            )

        return None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> Optional[bool]:
        if exc_val is not None:
            # These two exceptions are actually the same type under the hood--however, nothing
            # that I can find in the Python docs suggest that this HAS to be the case. Either
            # way, suppress cancellation errors that happen when we are closed, because .
            return self.allow() and (
                exc_type is asyncio.CancelledError or exc_type is concurrent.futures.CancelledError
            )

        return None


def _translate_exceptions(conn: "Connection") -> ExceptionTranslator:
    """
    Return an (async) context manager that translates exceptions thrown from low-level gRPC calls
    to high-level dazl exceptions.
    """
    return ExceptionTranslator(conn)


def _allow_cancel(allow: Callable[[], bool]) -> AllowCancellation:
    return AllowCancellation(allow)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file

import builtins as _builtins, sys, typing as _typing

from google.protobuf.descriptor import EnumDescriptor
from google.protobuf.message import Message as _Message

if sys.version_info >= (3, 8):
    from typing import Literal as _L
else:
    from typing_extensions import Literal as _L

__all__ = [
    "ExperimentalFeatures",
    "ExperimentalSelfServiceErrorCodes",
    "ExperimentalStaticTime",
    "CommandDeduplicationFeatures",
    "ExperimentalOptionalLedgerId",
    "CommandDeduplicationPeriodSupport",
    "ExperimentalContractIds",
    "ExperimentalCommitterEventLog",
]

class CommandDeduplicationType:
    DESCRIPTOR: _typing.ClassVar[EnumDescriptor] = ...
    ASYNC_ONLY: _typing.ClassVar[_L[0]] = ...
    ASYNC_AND_CONCURRENT_SYNC: _typing.ClassVar[_L[1]] = ...
ASYNC_ONLY = _L[0]
ASYNC_AND_CONCURRENT_SYNC = _L[1]


class ExperimentalFeatures(_Message):
    @property
    def self_service_error_codes(self) -> ExperimentalSelfServiceErrorCodes: ...
    @property
    def static_time(self) -> ExperimentalStaticTime: ...
    @property
    def command_deduplication(self) -> CommandDeduplicationFeatures: ...
    @property
    def optional_ledger_id(self) -> ExperimentalOptionalLedgerId: ...
    @property
    def contract_ids(self) -> ExperimentalContractIds: ...
    @property
    def committer_event_log(self) -> ExperimentalCommitterEventLog: ...
    def __init__(self, *, self_service_error_codes: _typing.Optional[ExperimentalSelfServiceErrorCodes] = ..., static_time: _typing.Optional[ExperimentalStaticTime] = ..., command_deduplication: _typing.Optional[CommandDeduplicationFeatures] = ..., optional_ledger_id: _typing.Optional[ExperimentalOptionalLedgerId] = ..., contract_ids: _typing.Optional[ExperimentalContractIds] = ..., committer_event_log: _typing.Optional[ExperimentalCommitterEventLog] = ...): ...
    def HasField(self, field_name: _L["self_service_error_codes", "static_time", "command_deduplication", "optional_ledger_id", "contract_ids", "committer_event_log"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["self_service_error_codes", "static_time", "command_deduplication", "optional_ledger_id", "contract_ids", "committer_event_log"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExperimentalSelfServiceErrorCodes(_Message):
    def __init__(self): ...
    def HasField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def ClearField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExperimentalStaticTime(_Message):
    supported: _builtins.bool
    def __init__(self, *, supported: _typing.Optional[_builtins.bool] = ...): ...
    def HasField(self, field_name: _L["supported"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["supported"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CommandDeduplicationFeatures(_Message):
    @property
    def deduplication_period_support(self) -> CommandDeduplicationPeriodSupport: ...
    @property
    def deduplication_type(self) -> _L[0, 1]: ...
    max_deduplication_duration_enforced: _builtins.bool
    def __init__(self, *, deduplication_period_support: _typing.Optional[CommandDeduplicationPeriodSupport] = ..., deduplication_type: _typing.Optional[_L['ASYNC_ONLY', 0, 'ASYNC_AND_CONCURRENT_SYNC', 1]] = ..., max_deduplication_duration_enforced: _typing.Optional[_builtins.bool] = ...): ...
    def HasField(self, field_name: _L["deduplication_period_support", "deduplication_type", "max_deduplication_duration_enforced"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["deduplication_period_support", "deduplication_type", "max_deduplication_duration_enforced"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExperimentalOptionalLedgerId(_Message):
    def __init__(self): ...
    def HasField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def ClearField(self, field_name: _typing.NoReturn) -> _typing.NoReturn: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class CommandDeduplicationPeriodSupport(_Message):
    @property
    def offset_support(self) -> _L[0, 1, 2]: ...
    @property
    def duration_support(self) -> _L[0, 1]: ...
    def __init__(self, *, offset_support: _typing.Optional[_L['OFFSET_NOT_SUPPORTED', 0, 'OFFSET_NATIVE_SUPPORT', 1, 'OFFSET_CONVERT_TO_DURATION', 2]] = ..., duration_support: _typing.Optional[_L['DURATION_NATIVE_SUPPORT', 0, 'DURATION_CONVERT_TO_OFFSET', 1]] = ...): ...
    def HasField(self, field_name: _L["offset_support", "duration_support"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["offset_support", "duration_support"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExperimentalContractIds(_Message):
    @property
    def v1(self) -> _L[0, 1]: ...
    def __init__(self, *, v1: _typing.Optional[_L['SUFFIXED', 0, 'NON_SUFFIXED', 1]] = ...): ...
    def HasField(self, field_name: _L["v1"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["v1"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...

class ExperimentalCommitterEventLog(_Message):
    @property
    def event_log_type(self) -> _L[0, 1]: ...
    def __init__(self, *, event_log_type: _typing.Optional[_L['CENTRALIZED', 0, 'DISTRIBUTED', 1]] = ...): ...
    def HasField(self, field_name: _L["event_log_type"]) -> _builtins.bool: ...
    def ClearField(self, field_name: _L["event_log_type"]) -> None: ...
    def WhichOneof(self, oneof_group: _typing.NoReturn) -> _typing.NoReturn: ...



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import gather
from os import PathLike
import sys
from typing import BinaryIO, Callable, Collection, List, Optional, Sequence, Union

from .. import connect
from ..ledger import PartyInfo
from ..ledger.aio import Connection
from ..prim import Party

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

__all__ = ["connect_with_new_party"]


NameGenFn = Callable[[int], Optional[str]]


def connect_with_new_party(
    *,
    party_count=1,
    url=None,
    read_as=None,
    act_as=None,
    admin=False,
    ledger_id=None,
    dar=None,
    identifier_hint=None,
    display_name=None,
) -> "ConnectWithNewParty":
    """
    A helper function for connecting to a ledger as a brand new party. This isn't normally useful
    outside of tests, where having creating new parties for each test helps keep test data isolated.

    Like the :func:`dazl.ledger.connect` function, all parameters are optional, and if unspecified,
    get their default values from the environment.

    This function can NOT be used with ledgers that require authentication since it has no way to
    ask a token granter for an appropriate token for the newly allocated party.

    :param url:
        The URL of the ledger to connect to. This ledger MUST NOT require authentication!
    :type url: str
    :param read_as:
        Extra parties to for which read-as access should be granted.
    :type read_as: str or Collection[str]
    :param act_as:
        Extra parties to for which act-as access should be granted. The created party is added to
        this list.
    :type read_as: str or Collection[str]
    :param admin:
        ``True`` if the created connection should have "admin" rights; otherwise ``False``. Note
        that in order to allocate parties, an initial connection with admin rights is ALWAYS
        created, and the value of this call only affects the returned connection.
    :type admin: bool
    :param ledger_id:
        The ledger ID to connect to. As with the ``connect`` api, if the target is the gRPC Ledger
        API, then supplying this field is optional, and with an HTTP JSON API server, this field
        must be supplied.
    :type ledger_id: str
    :param dar:
        A path to a file, the contents of the file (as ``bytes``), or a byte buffer that point to a
        DAR that should also be uploaded.
    :type dar: str or bytes or PathLike or BinaryIO
    :param party_count:
        The number of parties (and connections) to create.
    :type party_count: int
    :param identifier_hint:
        A hint to the backing participant. Note that the Daml ledger is free to ignore this
        hint entirely. If `party_count` is greater than 1, then
    :type identifier_hint: str
    :param display_name:
        A human-readable name that corresponds to the allocated ``Party``.
    :return:
        An asynchronous context manager that returns one or more :class:`ConnectionWithParty`
        objects.
    """
    if party_count < 0:
        raise ValueError("party_count must be a positive number")

    contents = None  # type: Optional[bytes]
    if dar is not None:
        if isinstance(dar, bytes):
            contents = dar
        elif isinstance(dar, (str, PathLike)):
            with open(dar, "rb") as buf:
                contents = buf.read()
        else:
            contents = dar.read()

    return ConnectWithNewParty(
        party_count=party_count,
        url=url,
        read_as=read_as,
        act_as=as_party_collection(act_as),
        admin=admin,
        ledger_id=ledger_id,
        dar=contents,
        identifier_hint_fn=convert_to_fn(identifier_hint, party_count=party_count),
        display_name_fn=convert_to_fn(display_name, party_count=party_count),
    )


class ConnectionWithParty:
    # We're cheating here; we declare the types of our fields as properties in the typing file,
    # but really they're just plain attributes
    # noinspection PyPropertyAccess
    def __init__(self, connection, party):
        self.connection = connection
        self.party = party

    def __getitem__(self, item: Literal[0]) -> "ConnectionWithParty":
        return self

    def __len__(self) -> Literal[1]:
        return 1


def as_party_collection(p: Union[None, Party, Collection[Party]]) -> Collection[Party]:
    if not p:
        return []
    elif isinstance(p, str):
        return [Party(o) for o in p.split(",")]
    else:
        return p


def convert_to_fn(party_or_fn: Union[None, str, NameGenFn], *, party_count: int = 1) -> NameGenFn:
    if party_or_fn is None:
        return lambda _: None
    if isinstance(party_or_fn, str):
        if party_count == 1:
            return lambda _: party_or_fn
        else:
            return lambda i: f"{party_or_fn}{i:03}"
    else:
        return party_or_fn


class ConnectWithNewParty:
    def __init__(
        self,
        *,
        url: Optional[str],
        read_as: Union[Party, Collection[Party]],
        act_as: Collection[Party],
        admin: Optional[bool],
        ledger_id: str,
        dar: Optional[bytes],
        party_count: int,
        identifier_hint_fn: NameGenFn,
        display_name_fn: NameGenFn,
    ):
        self.url = url
        self.read_as = read_as
        self.act_as = act_as
        self.admin = admin
        self.ledger_id = ledger_id
        self.dar = dar
        self.party_count = party_count
        self.identifier_hint_fn = identifier_hint_fn
        self.display_name_fn = display_name_fn
        self.connections = []  # type: List[Connection]

    async def __aenter__(self) -> "Sequence[ConnectionWithParty]":
        party_infos = await self._set_up_ledger()
        ret = []
        for party_info in party_infos:
            party_conn = connect(
                url=self.url,
                read_as=self.read_as,
                act_as=[party_info.party, *self.act_as],
                admin=self.admin,
                ledger_id=self.ledger_id,
            )
            await party_conn.open()
            ret.append(ConnectionWithParty(party_conn, party_info.party))
        if len(ret) == 1:
            return ret[0]
        else:
            return tuple(ret)

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await gather(*(conn.close() for conn in self.connections))

    async def _set_up_ledger(self) -> Sequence[PartyInfo]:
        # Party allocation and package uploading require "admin" access, so we disregard the given
        # user parameters to upload packages and allocate parties
        async with connect(url=self.url, admin=True, ledger_id=self.ledger_id) as admin_conn:
            if self.dar is not None:
                await admin_conn.upload_package(self.dar)

            return await gather(
                *(
                    admin_conn.allocate_party(
                        identifier_hint=self.identifier_hint_fn(i),
                        display_name=self.display_name_fn(i),
                    )
                    for i in range(self.party_count)
                )
            )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off
# isort: skip_file
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/commands.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import value_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_value__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%com/daml/ledger/api/v1/commands.proto\x12\x16\x63om.daml.ledger.api.v1\x1a\"com/daml/ledger/api/v1/value.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\"\xc8\x05\n\x08\x43ommands\x12\x1b\n\tledger_id\x18\x01 \x01(\tR\x08ledgerId\x12\x1f\n\x0bworkflow_id\x18\x02 \x01(\tR\nworkflowId\x12%\n\x0e\x61pplication_id\x18\x03 \x01(\tR\rapplicationId\x12\x1d\n\ncommand_id\x18\x04 \x01(\tR\tcommandId\x12\x14\n\x05party\x18\x05 \x01(\tR\x05party\x12;\n\x08\x63ommands\x18\x08 \x03(\x0b\x32\x1f.com.daml.ledger.api.v1.CommandR\x08\x63ommands\x12N\n\x12\x64\x65\x64uplication_time\x18\t \x01(\x0b\x32\x19.google.protobuf.DurationB\x02\x18\x01H\x00R\x11\x64\x65\x64uplicationTime\x12R\n\x16\x64\x65\x64uplication_duration\x18\x0f \x01(\x0b\x32\x19.google.protobuf.DurationH\x00R\x15\x64\x65\x64uplicationDuration\x12\x33\n\x14\x64\x65\x64uplication_offset\x18\x10 \x01(\tH\x00R\x13\x64\x65\x64uplicationOffset\x12I\n\x13min_ledger_time_abs\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x10minLedgerTimeAbs\x12H\n\x13min_ledger_time_rel\x18\x0b \x01(\x0b\x32\x19.google.protobuf.DurationR\x10minLedgerTimeRel\x12\x15\n\x06\x61\x63t_as\x18\x0c \x03(\tR\x05\x61\x63tAs\x12\x17\n\x07read_as\x18\r \x03(\tR\x06readAs\x12#\n\rsubmission_id\x18\x0e \x01(\tR\x0csubmissionIdB\x16\n\x14\x64\x65\x64uplication_periodJ\x04\x08\x06\x10\x07J\x04\x08\x07\x10\x08\"\xd4\x02\n\x07\x43ommand\x12?\n\x06\x63reate\x18\x01 \x01(\x0b\x32%.com.daml.ledger.api.v1.CreateCommandH\x00R\x06\x63reate\x12\x45\n\x08\x65xercise\x18\x02 \x01(\x0b\x32\'.com.daml.ledger.api.v1.ExerciseCommandH\x00R\x08\x65xercise\x12T\n\rexerciseByKey\x18\x04 \x01(\x0b\x32,.com.daml.ledger.api.v1.ExerciseByKeyCommandH\x00R\rexerciseByKey\x12`\n\x11\x63reateAndExercise\x18\x03 \x01(\x0b\x32\x30.com.daml.ledger.api.v1.CreateAndExerciseCommandH\x00R\x11\x63reateAndExerciseB\t\n\x07\x63ommand\"\x9f\x01\n\rCreateCommand\x12\x43\n\x0btemplate_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12I\n\x10\x63reate_arguments\x18\x02 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.RecordR\x0f\x63reateArguments\"\xd7\x01\n\x0f\x45xerciseCommand\x12\x43\n\x0btemplate_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12\x1f\n\x0b\x63ontract_id\x18\x02 \x01(\tR\ncontractId\x12\x16\n\x06\x63hoice\x18\x03 \x01(\tR\x06\x63hoice\x12\x46\n\x0f\x63hoice_argument\x18\x04 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0e\x63hoiceArgument\"\xfd\x01\n\x14\x45xerciseByKeyCommand\x12\x43\n\x0btemplate_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12@\n\x0c\x63ontract_key\x18\x02 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0b\x63ontractKey\x12\x16\n\x06\x63hoice\x18\x03 \x01(\tR\x06\x63hoice\x12\x46\n\x0f\x63hoice_argument\x18\x04 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0e\x63hoiceArgument\"\x8a\x02\n\x18\x43reateAndExerciseCommand\x12\x43\n\x0btemplate_id\x18\x01 \x01(\x0b\x32\".com.daml.ledger.api.v1.IdentifierR\ntemplateId\x12I\n\x10\x63reate_arguments\x18\x02 \x01(\x0b\x32\x1e.com.daml.ledger.api.v1.RecordR\x0f\x63reateArguments\x12\x16\n\x06\x63hoice\x18\x03 \x01(\tR\x06\x63hoice\x12\x46\n\x0f\x63hoice_argument\x18\x04 \x01(\x0b\x32\x1d.com.daml.ledger.api.v1.ValueR\x0e\x63hoiceArgumentB\x8c\x01\n\x16\x63om.daml.ledger.api.v1B\x12\x43ommandsOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\xaa\x02\x16\x43om.Daml.Ledger.Api.V1b\x06proto3')



_COMMANDS = DESCRIPTOR.message_types_by_name['Commands']
_COMMAND = DESCRIPTOR.message_types_by_name['Command']
_CREATECOMMAND = DESCRIPTOR.message_types_by_name['CreateCommand']
_EXERCISECOMMAND = DESCRIPTOR.message_types_by_name['ExerciseCommand']
_EXERCISEBYKEYCOMMAND = DESCRIPTOR.message_types_by_name['ExerciseByKeyCommand']
_CREATEANDEXERCISECOMMAND = DESCRIPTOR.message_types_by_name['CreateAndExerciseCommand']
Commands = _reflection.GeneratedProtocolMessageType('Commands', (_message.Message,), {
  'DESCRIPTOR' : _COMMANDS,
  '__module__' : 'com.daml.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Commands)
  })
_sym_db.RegisterMessage(Commands)

Command = _reflection.GeneratedProtocolMessageType('Command', (_message.Message,), {
  'DESCRIPTOR' : _COMMAND,
  '__module__' : 'com.daml.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.Command)
  })
_sym_db.RegisterMessage(Command)

CreateCommand = _reflection.GeneratedProtocolMessageType('CreateCommand', (_message.Message,), {
  'DESCRIPTOR' : _CREATECOMMAND,
  '__module__' : 'com.daml.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CreateCommand)
  })
_sym_db.RegisterMessage(CreateCommand)

ExerciseCommand = _reflection.GeneratedProtocolMessageType('ExerciseCommand', (_message.Message,), {
  'DESCRIPTOR' : _EXERCISECOMMAND,
  '__module__' : 'com.daml.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExerciseCommand)
  })
_sym_db.RegisterMessage(ExerciseCommand)

ExerciseByKeyCommand = _reflection.GeneratedProtocolMessageType('ExerciseByKeyCommand', (_message.Message,), {
  'DESCRIPTOR' : _EXERCISEBYKEYCOMMAND,
  '__module__' : 'com.daml.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.ExerciseByKeyCommand)
  })
_sym_db.RegisterMessage(ExerciseByKeyCommand)

CreateAndExerciseCommand = _reflection.GeneratedProtocolMessageType('CreateAndExerciseCommand', (_message.Message,), {
  'DESCRIPTOR' : _CREATEANDEXERCISECOMMAND,
  '__module__' : 'com.daml.ledger.api.v1.commands_pb2'
  # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.CreateAndExerciseCommand)
  })
_sym_db.RegisterMessage(CreateAndExerciseCommand)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.daml.ledger.api.v1B\022CommandsOuterClassZEgithub.com/digital-asset/dazl-client/v7/go/api/com/daml/ledger/api/v1\252\002\026Com.Daml.Ledger.Api.V1'
  _COMMANDS.fields_by_name['deduplication_time']._options = None
  _COMMANDS.fields_by_name['deduplication_time']._serialized_options = b'\030\001'
  _COMMANDS._serialized_start=167
  _COMMANDS._serialized_end=879
  _COMMAND._serialized_start=882
  _COMMAND._serialized_end=1222
  _CREATECOMMAND._serialized_start=1225
  _CREATECOMMAND._serialized_end=1384
  _EXERCISECOMMAND._serialized_start=1387
  _EXERCISECOMMAND._serialized_end=1602
  _EXERCISEBYKEYCOMMAND._serialized_start=1605
  _EXERCISEBYKEYCOMMAND._serialized_end=1858
  _CREATEANDEXERCISECOMMAND._serialized_start=1861
  _CREATEANDEXERCISECOMMAND._serialized_end=2127
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module has been relocated to ``dazl.client._conn_settings``.
"""

__all__ = ["HTTPConnectionSettings", "SSLSettings", "OAuthSettings", "connection_settings"]

from ..client._conn_settings import (
    HTTPConnectionSettings,
    OAuthSettings,
    SSLSettings,
    connection_settings,
)



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any
import warnings

from ..damlast.daml_lf_1 import TypeConName
from ..prim import ContractId
from ..protocols.serializers import AbstractSerializer
from ..values import CanonicalMapper
from .commands import CreateAndExerciseCommand, CreateCommand, ExerciseByKeyCommand, ExerciseCommand


class ValidateSerializer(AbstractSerializer):
    """
    A serializer that doesn't actually write to an on-wire form, but merely does lightweight
    conversion and give on-wire serialization the best chance of succeeding.

    This class essentially handles type conversion before a command is fully created and sent to
    a background thread for eventual writing to the Ledger API.
    """

    mapper = CanonicalMapper()

    def serialize_create_command(
        self, name: "TypeConName", template_args: "Any"
    ) -> "CreateCommand":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return CreateCommand(template=name, arguments=template_args)

    def serialize_exercise_command(
        self, contract_id: "ContractId", choice_name: str, choice_args: "Any"
    ) -> "ExerciseCommand":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return ExerciseCommand(contract=contract_id, choice=choice_name, arguments=choice_args)

    def serialize_exercise_by_key_command(
        self,
        template_name: "TypeConName",
        key_arguments: Any,
        choice_name: str,
        choice_arguments: "Any",
    ) -> "ExerciseByKeyCommand":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return ExerciseByKeyCommand(
                template=template_name,
                contract_key=key_arguments,
                choice=choice_name,
                choice_argument=choice_arguments,
            )

    def serialize_create_and_exercise_command(
        self,
        template_name: "TypeConName",
        create_arguments: Any,
        choice_name: str,
        choice_arguments: "Any",
    ) -> "CreateAndExerciseCommand":
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            return CreateAndExerciseCommand(
                template=template_name,
                arguments=create_arguments,
                choice=choice_name,
                choice_argument=choice_arguments,
            )



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Awaitable, Callable, Collection, Mapping, Optional, Sequence, TypeVar, Union

from ..prim import ContractData, ContractId, Party

__all__ = [
    "Event",
    "InitEvent",
    "ReadyEvent",
    "CreateEvent",
    "ArchiveEvent",
    "ExerciseResponse",
    "PartyInfo",
    "Query",
    "EventHandler",
    "AEventHandler",
]

Query = Union[None, Mapping[str, Any], Collection[Mapping[str, Any]]]


class Event:
    """
    Superclass of all dazl events.
    """

    __slots__ = ()

    def __setattr__(self, key, value):
        raise AttributeError("Event instances are read-only")


E = TypeVar("E", bound=Event)

EventHandler = Callable[[E], None]
AEventHandler = Union[Callable[[E], None], Callable[[E], Awaitable[None]]]


class InitEvent(Event):
    pass


class ReadyEvent(Event):
    pass


class CreateEvent(Event):
    __slots__ = ("_cid", "_cdata")

    _cid: "ContractId"
    _cdata: "ContractData"

    def __init__(self, cid: "ContractId", cdata: "ContractData"):
        object.__setattr__(self, "_cid", cid)
        object.__setattr__(self, "_cdata", cdata)

    @property
    def cid(self) -> "ContractId":
        return self._cid

    @property
    def cdata(self) -> "ContractData":
        return self._cdata


class ArchiveEvent(Event):
    __slots__ = ("_cid",)

    _cid: "ContractId"

    def __init__(self, cid: "ContractId"):
        object.__setattr__(self, "_cid", cid)

    @property
    def cid(self) -> "ContractId":
        return self._cid


class ExerciseResponse:
    __slots__ = "_result", "_events"

    _result: "Optional[Any]"
    _events: "Sequence[Union[CreateEvent, ArchiveEvent]]"

    def __init__(
        self, result: "Optional[Any]", events: "Sequence[Union[CreateEvent, ArchiveEvent]]"
    ):
        object.__setattr__(self, "_result", result)
        object.__setattr__(self, "_events", tuple(events))

    @property
    def result(self) -> "Optional[Any]":
        return self._result

    @property
    def events(self) -> "Sequence[Union[CreateEvent, ArchiveEvent]]":
        return self._events

    def __repr__(self):
        return f"ExerciseResponse(result={self.result}, events={self.events})"


class PartyInfo:
    __slots__ = "_party", "_display_name", "_is_local"

    _party: "Party"
    _display_name: str
    _is_local: bool

    def __init__(self, party: "Party", display_name: str, is_local: bool):
        object.__setattr__(self, "_party", party)
        object.__setattr__(self, "_display_name", display_name)
        object.__setattr__(self, "_is_local", is_local)

    @property
    def party(self) -> "Party":
        return self._party

    @property
    def display_name(self) -> str:
        return self._display_name

    @property
    def is_local(self) -> bool:
        return self._is_local



# Copyright (c) 2017-2022 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from asyncio import get_event_loop
from contextlib import ExitStack
from datetime import timedelta
import logging
import os
from pathlib import Path
import subprocess
import tempfile
import threading
from typing import Any, Dict, Mapping, Optional, Sequence, Union, cast

from ..util import ProcessLogger, find_free_port, kill_process_tree, wait_for_process_port
from ._cert import Certificate, cert_gen

__all__ = ["DEFAULT_TIMEOUT", "SandboxLauncher"]

DEFAULT_TIMEOUT = timedelta(seconds=30)


class SandboxLauncher:
    """
    A simple launcher for an in-memory Sandbox.
    """

    def __init__(
        self,
        *,
        project_root: "Union[None, str, os.PathLike]" = None,
        version: "Optional[str]" = None,
        timeout: "Optional[timedelta]" = DEFAULT_TIMEOUT,
        ledger_id: "Optional[str]" = None,
        use_auth: bool = False,
        use_tls: bool = False,
    ):
        self.log = logging.getLogger("sandbox")
        self._lock = threading.RLock()
        self._project_root = project_root
        self._version = version
        self._timeout = timeout
        self._ledger_id = ledger_id
        self._process = None  # type: Optional[subprocess.Popen]
        self._url = None  # type: Optional[str]
        self._exit_stack = ExitStack()
        self._use_auth = use_auth
        self._use_tls = use_tls

        self._certificate = None  # type: Optional[Certificate]
        self._crt_file = None  # type: Optional[Path]
        self._key_file = None  # type: Optional[Path]

        if self._use_auth or self._use_tls:
            if self._project_root is not None:
                raise RuntimeError(
                    "use_auth/use_tls cannot be specified when specifying a project root"
                )

        url = os.environ.get("DAZL_TEST_DAML_LEDGER_URL")
        if url:
            # if we're supposed to use a different already-running implementation,
            # then do that instead
            logging.info(
                "Using the sandbox at %s because `DAZL_TEST_DAML_LEDGER_URL` is defined", url
            )
            self._url = url

    @property
    def public_cert(self) -> bytes:
        """
        Return the public key corresponding to the private key that was used to start the
        sandbox using TLS.
        """
        if self._certificate is None:
            raise RuntimeError("auth/tls was not enabled, so a public cert was not generated")

        return self._certificate.public_cert

    def sign_token(self, claims: "Mapping[str, Any]") -> str:
        """
        Sign a token using the private key that the sandbox was launched with.
        """
        if self._certificate is None:
            raise RuntimeError("this sandbox was not started with auth")

        import jwt

        return jwt.encode(
            # there is a bug in the jwt typing rules that falsely state the claims are
            # Dict[str, Any] when Mapping[str, Any] would actually work
            cast(Dict[str, Any], claims),
            self._certificate.private_key.decode("utf-8"),
            algorithm="RS256",
        )

    @property
    def url(self) -> str:
        """
        Return a URL that can be used to connect to this running Sandbox.
        """
        u = self._url
        if u is not None:
            return u
        else:
            raise RuntimeError("the sandbox is not running (did you call start()?)")

    def start(self) -> None:
        """
        Start the sandbox and return a URL that can be used to connect to it.
        """
        if self._url is not None:
            return

        port = find_free_port()

        env = os.environ.copy()

        # Running dazl's tests against a different Sandbox merely requires the DAML_SDK_VERSION
        # variable be set to a different value
        if self._version is not None:
            env["DAML_SDK_VERSION"] = self._version
            major_version_str, _, _ = self._version.partition(".")
            major_version = int(major_version_str)
        else:
            # TODO: interrogate the local daml binary to figure out what version is defaulted;
            #  assume 1 for now
            major_version = 1

        if self._use_auth or self._use_tls:
            # We're supposed to launch with auth and/or TLS.
            # Both of these flags moved around from Daml 1.x to Daml 2.x,
            # so determine which set of flags to use.
            if self._version is None:
                raise RuntimeError("when tls/auth is on, a version must be specified")

        with self._lock:
            # one last check inside the lock
            if self._url is not None:
                return

            if self._process is not None:
                # this shouldn't ever happen, but we explicitly check for it because running
                # multiple sandbox processes could be very confusing
                raise RuntimeError("we somehow have a process without a URL")

            if self._use_auth or self._use_tls:
                self._certificate = cert_gen(subject_alternative_name="localhost")

                temp_dir = Path(self._exit_stack.enter_context(tempfile.TemporaryDirectory()))
                self._crt_file = temp_dir / "tmp.crt"
                self._key_file = temp_dir / "tmp.key"

                self._crt_file.write_bytes(self._certificate.public_cert)
                self._key_file.write_bytes(self._certificate.private_key)

            options = SandboxOptions(
                port=port,
                ledger_id=self._ledger_id,
                project_root=self._project_root,  # type: ignore
                use_auth=self._use_auth,
                use_tls=self._use_tls,
                cert_file=self._crt_file,
                key_file=self._key_file,
            )
            cmdline = options.daml_cmdline(major_version)

            self.log.info("Launching sandbox: %s", cmdline)

            self._process = subprocess.Popen(
                cmdline,
                env=env,
                cwd=self._project_root,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            ProcessLogger(self._process, self.log).start()
            wait_for_process_port(self._process, port, DEFAULT_TIMEOUT)

            self._url = f"http://localhost:{port}"

    def stop(self) -> None:
        """
        Stop the sandbox.
        """
        process = self._process
        if process is not None:
            # Clean up the process that we started. Note that some really old versions of the SDK
            # had issues that left dangling child processes running even after the parent process is
            # killed, so make sure that we find and destroy them too if the parent process doesn't
            # kill its own children quickly enough.
            try:
                kill_process_tree(process)
            finally:
                self._url = None
        self._exit_stack.close()

    def __enter__(self) -> "SandboxLauncher":
        """
        Start the sandbox.
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Stop the sandbox.
        """
        self.stop()

    async def __aenter__(self) -> "SandboxLauncher":
        """
        Start the sandbox.
        """
        loop = get_event_loop()
        await loop.run_in_executor(None, self.start)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Stop the sandbox.
        """
        loop = get_event_loop()
        await loop.run_in_executor(None, self.stop)


class SandboxOptions:
    port: int
    project_root: Optional[str] = None
    ledger_id: Optional[str] = None
    use_auth: bool = False
    use_tls: bool = False
    cert_file: Optional[Path] = None
    key_file: Optional[Path] = None

    def __init__(
        self,
        *,
        port: int,
        project_root: Optional[str] = None,
        ledger_id: Optional[str] = None,
        use_auth: bool = False,
        use_tls: bool = False,
        cert_file: Optional[Path] = None,
        key_file: Optional[Path] = None,
    ):
        object.__setattr__(self, "port", port)
        object.__setattr__(self, "project_root", project_root)
        object.__setattr__(self, "ledger_id", ledger_id)
        object.__setattr__(self, "use_auth", use_auth)
        object.__setattr__(self, "use_tls", use_tls)
        object.__setattr__(self, "cert_file", cert_file)
        object.__setattr__(self, "key_file", key_file)

    def daml_cmdline(self, __major_version: int) -> "Sequence[str]":
        if __major_version <= 1:
            return self._daml1_cmdline()
        elif __major_version == 2:
            return self._daml2_cmdline()
        else:
            raise ValueError("unknown Daml version")

    def _daml1_cmdline(self) -> Sequence[str]:
        if self.project_root:
            cmdline = ["daml", "start"]
            cmdline.extend(["--start-navigator=no", "--open-browser=no"])
            cmdline.append(f"--sandbox-port={self.port}")
        else:
            cmdline = ["daml", "sandbox", "--port", str(self.port)]

        if self.ledger_id:
            cmdline.extend(["--ledgerid", self.ledger_id])
        if self.use_auth:
            cmdline.extend(["--auth-jwt-rs256-crt", str(self.cert_file)])
        if self.use_tls:
            cmdline.extend(
                ["--crt", str(self.cert_file), "--pem", str(self.key_file), "--client-auth", "none"]
            )

        return cmdline

    def _daml2_cmdline(self) -> Sequence[str]:
        if self.ledger_id and (self.ledger_id != "sandbox"):
            raise ValueError('for Daml 2.x ledgers, ledger ID must be unset, or set to "sandbox"')

        participant_admin_port = find_free_port()
        domain_api_port = find_free_port()
        domain_admin_port = find_free_port()

        participant = "canton.participants.sandbox"
        domain = "canton.domains.local"

        cmdline = ["daml", "sandbox", "--port", str(self.port)]

        # for the time being, we don't use any of Canton's other ports, but it's important that
        # they are bound to random ports to avoid conflicts
        config_options = {
            f"{participant}.admin-api.port": str(participant_admin_port),
            f"{domain}.public-api.port": str(domain_api_port),
            f"{domain}.admin-api.port": str(domain_admin_port),
        }

        if self.use_auth:
            if self.cert_file is None:
                raise ValueError("cert_file must be specified when use_auth=True")

            config_options[f"{participant}.ledger-api.auth-services.0.type"] = "jwt-rs-256-crt"
            config_options[f"{participant}.ledger-api.auth-services.0.certificate"] = str(
                self.cert_file
            )
        if self.use_tls:
            # config_options["canton.participants.sandbox.ledger-api.host"] = "localhost"
            if self.cert_file is None:
                raise ValueError("cert_file must be specified when use_auth=True")
            if self.key_file is None:
                raise ValueError("cert_file must be specified when use_auth=True")

            config_options[f"{participant}.ledger-api.tls.cert-chain-file"] = str(self.cert_file)
            config_options[f"{participant}.ledger-api.tls.private-key-file"] = str(self.key_file)

        if config_options:
            cmdline.append("-C")
            cmdline.append(",".join(f"{key}={value}" for key, value in config_options.items()))

        return cmdline
